import numpy as np
import pytest

from semhash import SemHash
from semhash.datamodels import FilterResult
from semhash.utils import Encoder


def test_single_dataset_deduplication(use_ann: bool, model: Encoder) -> None:
    """Test single dataset deduplication."""
    # No duplicates
    texts = [
        "It's dangerous to go alone!",
        "The master sword can seal the darkness.",
        "Ganondorf has invaded Hyrule!",
    ]
    semhash = SemHash.from_records(records=texts, use_ann=use_ann, model=model)
    deduplicated_texts = semhash.self_deduplicate().selected

    assert deduplicated_texts == texts

    # With duplicates
    texts = [
        "It's dangerous to go alone!",
        "It's dangerous to go alone!",  # Exact duplicate
        "It's not safe to go alone!",  # Semantically similar
    ]
    semhash = SemHash.from_records(records=texts, use_ann=use_ann, model=model)
    deduplicated_texts = semhash.self_deduplicate(0.7).selected
    assert deduplicated_texts == ["It's dangerous to go alone!"]


def test_multi_dataset_deduplication(use_ann: bool, model: Encoder) -> None:
    """Test deduplication across two datasets."""
    # No duplicates
    texts1 = [
        "It's dangerous to go alone!",
        "It's a secret to everybody.",
        "Ganondorf has invaded Hyrule!",
    ]
    texts2 = [
        "Link is the hero of time.",
        "Zelda is the princess of Hyrule.",
        "Ganon is the king of thieves.",
    ]
    semhash = SemHash.from_records(texts1, columns=None, use_ann=use_ann, model=model)
    deduplicated_texts = semhash.deduplicate(texts2).selected
    assert deduplicated_texts == texts2

    # With duplicates
    texts2 = [
        "It's dangerous to go alone!",  # Exact duplicate
        "It's risky to go alone!",  # Semantically similar
        "Ganondorf has attacked Hyrule!",  # Semantically similar
    ]
    deduplicated_texts = semhash.deduplicate(texts2, threshold=0.7).selected
    assert deduplicated_texts == []


def test_single_dataset_deduplication_multicolumn(use_ann: bool, model: Encoder) -> None:
    """Test single dataset deduplication with multi-column records."""
    records = [
        {"question": "What is the hero's name?", "context": "The hero is Link", "answer": "Link"},
        {"question": "What is the hero's name?", "context": "The hero is Link", "answer": "Link"},  # Exact duplicate
        {
            "question": "Who is the protagonist?",
            "context": "In this story, Link is the hero",
            "answer": "Link",
        },  # Semantically similar
        {"question": "Who is the princess?", "context": "The princess is Zelda", "answer": "Zelda"},
    ]
    semhash = SemHash.from_records(
        records,
        columns=["question", "context", "answer"],
        use_ann=use_ann,
        model=model,
    )
    deduplicated = semhash.self_deduplicate(threshold=0.7)

    assert deduplicated.selected == [
        {"question": "What is the hero's name?", "context": "The hero is Link", "answer": "Link"},
        {"question": "Who is the princess?", "context": "The princess is Zelda", "answer": "Zelda"},
    ]


def test_multi_dataset_deduplication_multicolumn(use_ann: bool, model: Encoder) -> None:
    """Test multi dataset deduplication with multi-column records."""
    train_records = [
        {"question": "What is the hero's name?", "context": "The hero is Link", "answer": "Link"},
        {"question": "Who is the princess?", "context": "The princess is Zelda", "answer": "Zelda"},
    ]
    test_records = [
        {"question": "What is the hero's name?", "context": "The hero is Link", "answer": "Link"},  # Exact duplicate
        {
            "question": "Who is the princess?",
            "context": "Zelda is the princess",
            "answer": "Zelda",
        },  # Semantically similar
        {"question": "What is the villain's name?", "context": "The villain is Ganon", "answer": "Ganon"},
    ]
    semhash = SemHash.from_records(
        train_records,
        columns=["question", "context", "answer"],
        use_ann=use_ann,
        model=model,
    )
    deduplicated = semhash.deduplicate(test_records).selected
    assert deduplicated == [
        {"question": "What is the villain's name?", "context": "The villain is Ganon", "answer": "Ganon"}
    ]


def test_from_records_without_columns(use_ann: bool, model: Encoder) -> None:
    """Test fitting without specifying columns."""
    records = [
        {"question": "What is the hero's name?", "context": "The hero is Link", "answer": "Link"},
        {"question": "Who is the princess?", "context": "The princess is Zelda", "answer": "Zelda"},
    ]
    with pytest.raises(ValueError):
        SemHash.from_records(records, columns=None, use_ann=use_ann, model=model)


def test_deduplicate_with_only_exact_duplicates(use_ann: bool, model: Encoder) -> None:
    """Test deduplicating with only exact duplicates."""
    texts1 = [
        "It's dangerous to go alone!",
        "It's dangerous to go alone!",
        "It's dangerous to go alone!",
    ]
    texts2 = [
        "It's dangerous to go alone!",
        "It's dangerous to go alone!",
        "It's dangerous to go alone!",
    ]
    semhash = SemHash.from_records(texts1, use_ann=use_ann, model=model)
    deduplicated = semhash.self_deduplicate()
    assert deduplicated.selected == ["It's dangerous to go alone!"]

    deduplicated = semhash.deduplicate(texts2)
    assert deduplicated.selected == []


def test_self_find_representative(use_ann: bool, model: Encoder, train_texts: list[str]) -> None:
    """Test the self_find_representative method."""
    semhash = SemHash.from_records(records=train_texts, use_ann=use_ann, model=model)
    result = semhash.self_find_representative(
        candidate_limit=5,
        selection_size=3,
        lambda_param=0.5,
    )
    assert len(result.selected) == 3, "Expected 3 representatives"
    selected = {r["text"] for r in result.selected}
    assert selected == {
        "blueberry",
        "pineapple",
        "grape",
    }, "Expected representatives to be blueberry, pineapple, and grape"


def test_find_representative(use_ann: bool, model: Encoder, train_texts: list[str], test_texts: list[str]) -> None:
    """Test the find_representative method."""
    semhash = SemHash.from_records(records=train_texts, use_ann=use_ann, model=model)
    result = semhash.find_representative(records=test_texts, candidate_limit=5, selection_size=3, lambda_param=0.5)
    assert len(result.selected) == 3, "Expected 3 representatives"
    selected = {r["text"] for r in result.selected}
    assert selected == {"grapefruit", "banana", "apple"}, "Expected representatives to be grapefruit, banana, and apple"


def test_filter_outliers(use_ann: bool, model: Encoder, train_texts: list[str], test_texts: list[str]) -> None:
    """Test the filter_outliers method."""
    semhash = SemHash.from_records(records=train_texts, use_ann=use_ann, model=model)
    result = semhash.filter_outliers(records=test_texts, outlier_percentage=0.2)
    assert len(result.filtered) == 2, "Expected 2 outliers"
    assert len(result.selected) == len(test_texts) - 2
    filtered = {r["text"] for r in result.filtered}
    assert filtered == {"motorcycle", "plane"}, "Expected outliers to be motorcycle and plane"


def test_self_filter_outliers(use_ann: bool, model: Encoder, train_texts: list[str]) -> None:
    """Test the self_filter_outliers method."""
    semhash = SemHash.from_records(records=train_texts, use_ann=use_ann, model=model)
    result = semhash.self_filter_outliers(outlier_percentage=0.1)
    assert len(result.filtered) == 2, "Expected 2 outliers"
    assert len(result.selected) == len(train_texts) - 2
    filtered = {r["text"] for r in result.filtered}
    assert filtered == {"car", "bicycle"}, "Expected outliers to be car and bicycle"


def test__mmr(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test the _mmr method."""
    # Create a dummy SemHash instance
    semhash = SemHash(index=None, model=None, columns=["text"], was_string=True)  # type: ignore
    # Prepare a fake ranking with three records
    records = ["a", "b", "c"]
    scores = [3.0, 2.0, 1.0]
    ranking = FilterResult(selected=records, filtered=[], scores_selected=scores, scores_filtered=[])
    # Create dummy embeddings for the records
    embeddings = np.array([[1.0, 0.0], [0.5, 0.5], [0.0, 1.0]])
    # Monkeypatch featurize to return the dummy embeddings
    monkeypatch.setattr(semhash, "_featurize", lambda records, columns, model: embeddings)

    # Test lambda=1.0: pure relevance, should pick top 2 by score
    result_rel = semhash._mmr(ranking, candidate_limit=3, selection_size=2, lambda_param=1.0)
    assert result_rel.selected == ["a", "b"]

    # Test lambda=0.0: pure diversity, should first pick 'a', then pick most dissimilar: 'c'
    result_div = semhash._mmr(ranking, candidate_limit=3, selection_size=2, lambda_param=0.0)
    assert result_div.selected == ["a", "c"]


def test_mmr_invalid_lambda_raises() -> None:
    """Test that invalid lambda values raise ValueError."""
    semhash = SemHash(index=None, model=None, columns=["text"], was_string=True)  # type: ignore
    dummy = FilterResult(selected=["x"], filtered=[], scores_selected=[0.5], scores_filtered=[])
    with pytest.raises(ValueError):
        semhash._mmr(dummy, candidate_limit=1, selection_size=1, lambda_param=-0.1)
    with pytest.raises(ValueError):
        semhash._mmr(dummy, candidate_limit=1, selection_size=1, lambda_param=1.1)



import pytest
from model2vec import StaticModel


@pytest.fixture
def model() -> StaticModel:
    """Load a model for testing."""
    return StaticModel.from_pretrained("tests/data/test_model")


@pytest.fixture(params=[True, False], ids=["use_ann=True", "use_ann=False"])
def use_ann(request: pytest.FixtureRequest) -> bool:
    """Whether to use approximate nearest neighbors or not."""
    return request.param


@pytest.fixture
def train_texts() -> list[str]:
    """A list of train texts for testing outlier and representative filtering."""
    return [
        "apple",
        "banana",
        "cherry",
        "strawberry",
        "blueberry",
        "raspberry",
        "blackberry",
        "peach",
        "plum",
        "grape",
        "mango",
        "papaya",
        "pineapple",
        "watermelon",
        "orange",
        "lemon",
        "lime",
        "tangerine",
        "car",  # Outlier
        "bicycle",  # Outlier
    ]


@pytest.fixture
def test_texts() -> list[str]:
    """A list of test texts for testing outlier and representative filtering."""
    return [
        "apple",
        "banana",
        "kiwi",
        "fig",
        "apricot",
        "grapefruit",
        "pomegranate",
        "motorcycle",  # Outlier
        "plane",  # Outlier
    ]



import pytest

import semhash
import semhash.version
from semhash.datamodels import DeduplicationResult, DuplicateRecord, SelectedWithDuplicates


def test_deduplication_scoring() -> None:
    """Test the deduplication scoring."""
    d = DeduplicationResult(
        ["a", "b", "c"],
        [DuplicateRecord("a", False, [("b", 0.9)]), DuplicateRecord("b", False, [("c", 0.8)])],
        0.8,
    )
    assert d.duplicate_ratio == 0.4


def test_deduplication_scoring_exact() -> None:
    """Test the deduplication scoring."""
    d = DeduplicationResult(
        ["a", "b", "c"],
        [DuplicateRecord("a", True, [("b", 0.9)]), DuplicateRecord("b", False, [("c", 0.8)])],
        0.8,
    )
    assert d.exact_duplicate_ratio == 0.2


def test_deduplication_scoring_exact_empty() -> None:
    """Test the deduplication scoring."""
    d = DeduplicationResult([], [], 0.8, columns=["text"])
    assert d.exact_duplicate_ratio == 0.0


def test_deduplication_scoring_empty() -> None:
    """Test the deduplication scoring."""
    d = DeduplicationResult([], [], 0.8, columns=["text"])
    assert d.duplicate_ratio == 0.0


def test_rethreshold() -> None:
    """Test rethresholding the duplicates."""
    d = DuplicateRecord("a", False, [("b", 0.9), ("c", 0.8)])
    d._rethreshold(0.85)
    assert d.duplicates == [("b", 0.9)]


def test_rethreshold_empty() -> None:
    """Test rethresholding the duplicates."""
    d = DuplicateRecord("a", False, [])
    d._rethreshold(0.85)
    assert d.duplicates == []


def test_get_least_similar_from_duplicates() -> None:
    """Test getting the least similar duplicates."""
    d = DeduplicationResult(
        ["a", "b", "c"],
        [DuplicateRecord("a", False, [("b", 0.9), ("c", 0.7)]), DuplicateRecord("b", False, [("c", 0.8)])],
        0.8,
    )
    result = d.get_least_similar_from_duplicates(1)
    assert result == [("a", "c", 0.7)]


def test_get_least_similar_from_duplicates_empty() -> None:
    """Test getting the least similar duplicates."""
    d = DeduplicationResult([], [], 0.8, columns=["text"])
    assert d.get_least_similar_from_duplicates(1) == []


def test_rethreshold_deduplication_result() -> None:
    """Test rethresholding the duplicates."""
    d = DeduplicationResult(
        ["a", "b", "c"],
        [
            DuplicateRecord("d", False, [("x", 0.9), ("y", 0.8)]),
            DuplicateRecord("e", False, [("z", 0.8)]),
        ],
        0.8,
    )
    d.rethreshold(0.85)
    assert d.filtered == [DuplicateRecord("d", False, [("x", 0.9)])]
    assert d.selected == ["a", "b", "c", "e"]


def test_rethreshold_exception() -> None:
    """Test rethresholding throws an exception."""
    d = DeduplicationResult(
        ["a", "b", "c"],
        [
            DuplicateRecord("d", False, [("x", 0.9), ("y", 0.8)]),
            DuplicateRecord("e", False, [("z", 0.8)]),
        ],
        0.7,
    )
    with pytest.raises(ValueError):
        d.rethreshold(0.6)


def test_deprecation_deduplicated_duplicates() -> None:
    """Test deprecation warnings for deduplicated and duplicates fields."""
    if semhash.version.__version__ < "0.4.0":
        with pytest.warns(DeprecationWarning):
            d = DeduplicationResult(
                deduplicated=["a", "b", "c"],
                duplicates=[
                    DuplicateRecord("d", False, [("x", 0.9), ("y", 0.8)]),
                    DuplicateRecord("e", False, [("z", 0.8)]),
                ],
                threshold=0.8,
            )
    else:
        raise ValueError("deprecate `deduplicated` and `duplicates` fields in `DeduplicationResult`")
    assert d.selected == ["a", "b", "c"]
    assert d.filtered == [
        DuplicateRecord("d", False, [("x", 0.9), ("y", 0.8)]),
        DuplicateRecord("e", False, [("z", 0.8)]),
    ]


def test_selected_with_duplicates_strings() -> None:
    """Test selected_with_duplicates for strings."""
    d = DeduplicationResult(
        selected=["original"],
        filtered=[
            DuplicateRecord("duplicate_1", False, [("original", 0.9)]),
            DuplicateRecord("duplicate_2", False, [("original", 0.8)]),
        ],
        threshold=0.8,
    )

    expected = [
        SelectedWithDuplicates(
            record="original",
            duplicates=[("duplicate_1", 0.9), ("duplicate_2", 0.8)],
        )
    ]
    assert d.selected_with_duplicates == expected


def test_selected_with_duplicates_dicts() -> None:
    """Test selected_with_duplicates for dicts."""
    selected = {"id": 0, "text": "hello"}
    d = DeduplicationResult(
        selected=[selected],
        filtered=[
            DuplicateRecord({"id": 1, "text": "hello"}, True, [(selected, 1.0)]),
            DuplicateRecord({"id": 2, "text": "helllo"}, False, [(selected, 0.1)]),
        ],
        threshold=0.8,
        columns=["text"],
    )

    items = d.selected_with_duplicates
    assert len(items) == 1
    kept = items[0].record
    dups = items[0].duplicates
    assert kept == selected
    assert {r["id"] for r, _ in dups} == {1, 2}


def test_selected_with_duplicates_multi_column() -> None:
    """Test selected_with_duplicates for multi-columns."""
    selected = {"text": "hello", "text2": "world"}
    d = DeduplicationResult(
        selected=[selected],
        filtered=[
            DuplicateRecord({"text": "hello", "text2": "world"}, True, [(selected, 1.0)]),
            DuplicateRecord({"text": "helllo", "text2": "world"}, False, [(selected, 0.1)]),
        ],
        threshold=0.8,
        columns=["text", "text2"],
    )

    items = d.selected_with_duplicates
    assert len(items) == 1
    kept = items[0].record
    assert kept == selected


def test_selected_with_duplicates_unhashable_values() -> None:
    """Test selected_with_duplicates with unhashable values in records."""
    selected = {"text": "hello", "a": [1, 2, 3]}  # list -> unhashable value
    filtered = {"text": "hello", "a": [1, 2, 3], "flag": True}

    d = DeduplicationResult(
        selected=[selected],
        filtered=[DuplicateRecord(filtered, exact=False, duplicates=[(selected, 1.0)])],
        threshold=0.8,
        columns=["text"],
    )

    items = d.selected_with_duplicates
    assert items == [SelectedWithDuplicates(record=selected, duplicates=[(filtered, 1.0)])]


def test_selected_with_duplicates_removes_internal_duplicates() -> None:
    """Test that selected_with_duplicates removes internal duplicates that have the same hash."""
    selected = {"id": 0, "text": "hello"}
    filtered = {"id": 1, "text": "hello"}

    d = DeduplicationResult(
        selected=[selected],
        filtered=[
            DuplicateRecord(filtered, exact=False, duplicates=[(selected, 0.95)]),
            DuplicateRecord(filtered, exact=False, duplicates=[(selected, 0.90)]),
        ],
        threshold=0.8,
        columns=["text"],
    )

    items = d.selected_with_duplicates
    assert len(items) == 1

    selected_record = items[0].record
    duplicate_list = items[0].duplicates
    # Should keep the kept record unchanged
    assert selected_record == selected
    # The duplicate row must appear only once
    assert len(duplicate_list) == 1
    assert duplicate_list[0][0] == filtered



import json
import logging
from time import perf_counter

from datasets import load_dataset
from model2vec import StaticModel

from benchmarks.data import DATASET_DICT
from semhash import SemHash

# Set up logging
logger = logging.getLogger(__name__)


def main() -> None:  # noqa: C901
    """Run the benchmarks."""
    # Prepare lists to hold benchmark results
    train_dedup_results = []
    train_test_dedup_results = []
    # Load the model and initialize SemHash

    model = StaticModel.from_pretrained("minishlab/potion-base-8m")

    for dataset_name, record in DATASET_DICT.items():
        logger.info(f"Loading dataset: {dataset_name} from {record.name}")

        # Load train and test splits
        if record.sub_directory:
            train_ds = load_dataset(record.name, record.sub_directory, split=record.split_one)
            test_ds = load_dataset(record.name, record.sub_directory, split=record.split_two)
        else:
            train_ds = load_dataset(record.name, split=record.split_one)
            test_ds = load_dataset(record.name, split=record.split_two)

        # If the dataset has columns, use them
        if record.columns:
            columns = record.columns
            train_records = [dict(row) for row in train_ds]
            test_records = [dict(row) for row in test_ds]
        # Else, use the text_name
        else:
            train_records = train_ds[record.text_name]
            test_records = test_ds[record.text_name]
            columns = None

        # Build the SemHash instance
        build_start = perf_counter()
        semhash = SemHash.from_records(model=model, use_ann=True, records=train_records, columns=columns)
        build_end = perf_counter()
        build_time = build_end - build_start
        # Time how long it takes to deduplicate the train set
        train_only_start = perf_counter()
        deduplicated_train = semhash.self_deduplicate()
        train_only_end = perf_counter()

        train_only_dedup_time = train_only_end - train_only_start
        original_train_size = len(train_records)
        dedup_train_size = len(deduplicated_train.selected)

        percent_removed_train = deduplicated_train.duplicate_ratio * 100
        train_dedup_results.append(
            {
                "dataset": dataset_name,
                "original_train_size": original_train_size,
                "deduplicated_train_size": dedup_train_size,
                "percent_removed": percent_removed_train,
                "build_time_seconds": build_time,
                "deduplication_time_seconds": train_only_dedup_time,
                "time_seconds": train_only_dedup_time + build_time,
            }
        )

        logger.info(
            f"[TRAIN DEDUPLICATION] Dataset: {dataset_name}\n"
            f" - Original Train Size: {original_train_size}\n"
            f" - Deduplicated Train Size: {dedup_train_size}\n"
            f" - % Removed: {percent_removed_train:.2f}\n"
            f" - Deduplication Time (seconds): {train_only_dedup_time:.2f}\n"
            f" - Build Time (seconds): {build_time:.2f}\n"
            f" - Total Time (seconds): {train_only_dedup_time + build_time:.2f}\n"
        )

        # Time how long it takes to deduplicate the test set
        train_test_start = perf_counter()
        deduplicated_test = semhash.deduplicate(
            records=test_records,
        )
        train_test_end = perf_counter()
        train_test_dedup_time = train_test_end - train_test_start
        original_test_size = len(test_records)
        deduped_test_size = len(deduplicated_test.selected)
        percent_removed_test = deduplicated_test.duplicate_ratio * 100

        train_test_dedup_results.append(
            {
                "dataset": dataset_name,
                "train_size": original_train_size,
                "test_size": original_test_size,
                "deduplicated_test_size": deduped_test_size,
                "percent_removed": percent_removed_test,
                "build_time_seconds": build_time,
                "deduplication_time_seconds": train_test_dedup_time,
                "time_seconds": train_test_dedup_time + build_time,
            }
        )

        logger.info(
            f"[TRAIN/TEST DEDUPLICATION] Dataset: {dataset_name}\n"
            f" - Train Size: {original_train_size}\n"
            f" - Test Size: {original_test_size}\n"
            f" - Deduplicated Test Size: {deduped_test_size}\n"
            f" - % Removed: {percent_removed_test:.2f}\n"
            f" - Deduplication Time (seconds): {train_test_dedup_time:.2f}\n"
            f" - Build Time (seconds): {build_time:.2f}\n"
            f" - Total Time (seconds): {train_test_dedup_time + build_time:.2f}\n"
        )

    # Write the results to JSON files
    with open("benchmarks/results/train_benchmark_results.json", "w", encoding="utf-8") as f:
        json.dump(train_dedup_results, f, ensure_ascii=False, indent=2)

    with open("benchmarks/results/train_test_benchmark_results.json", "w", encoding="utf-8") as f:
        json.dump(train_test_dedup_results, f, ensure_ascii=False, indent=2)

    # Print the train table
    print("### Train Deduplication Benchmark\n")  # noqa T201
    print(  # noqa T201
        f"| {'Dataset':<20} | {'Original Train Size':>20} | {'Deduplicated Train Size':>24} | {'% Removed':>10} | {'Deduplication Time (s)':>24} |"
    )  # noqa T201
    print("|" + "-" * 22 + "|" + "-" * 22 + "|" + "-" * 26 + "|" + "-" * 12 + "|" + "-" * 26 + "|")  # noqa T201
    for r in train_dedup_results:
        print(  # noqa T201
            f"| {r['dataset']:<20} "
            f"| {r['original_train_size']:>20} "
            f"| {r['deduplicated_train_size']:>24} "
            f"| {r['percent_removed']:>10.2f} "
            f"| {r['time_seconds']:>24.2f} |"
        )

    print("\n")  # noqa T201

    # Print the train/test table
    print("### Train/Test Deduplication Benchmark\n")  # noqa T201
    print(  # noqa T201
        f"| {'Dataset':<20} | {'Train Size':>12} | {'Test Size':>12} | {'Deduplicated Test Size':>24} | {'% Removed':>10} | {'Deduplication Time (s)':>24} |"
    )  # noqa T201
    print("|" + "-" * 22 + "|" + "-" * 14 + "|" + "-" * 14 + "|" + "-" * 26 + "|" + "-" * 12 + "|" + "-" * 26 + "|")  # noqa T201
    for r in train_test_dedup_results:
        print(  # noqa T201
            f"| {r['dataset']:<20} "
            f"| {r['train_size']:>12} "
            f"| {r['test_size']:>12} "
            f"| {r['deduplicated_test_size']:>24} "
            f"| {r['percent_removed']:>10.2f} "
            f"| {r['time_seconds']:>24.2f} |"
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()



from __future__ import annotations

import json
import warnings
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Generic, Hashable, Sequence, TypeVar

from frozendict import frozendict
from typing_extensions import TypeAlias

from semhash.utils import to_frozendict

Record = TypeVar("Record", str, dict[str, Any])
DuplicateList: TypeAlias = list[tuple[Record, float]]


@dataclass
class DuplicateRecord(Generic[Record]):
    """
    A single record with its duplicates.

    Attributes
    ----------
        record: The original record being deduplicated.
        exact: Whether the record was identified as an exact match.
        duplicates: List of tuples consisting of duplicate records and their associated scores.

    """

    record: Record
    exact: bool
    duplicates: DuplicateList = field(default_factory=list)

    def _rethreshold(self, threshold: float) -> None:
        """Rethreshold the duplicates."""
        self.duplicates = [(d, score) for d, score in self.duplicates if score >= threshold]


@dataclass
class SelectedWithDuplicates(Generic[Record]):
    """
    A record that has been selected along with its duplicates.

    Attributes
    ----------
        record: The original record being selected.
        duplicates: List of tuples consisting of duplicate records and their associated scores.

    """

    record: Record
    duplicates: DuplicateList = field(default_factory=list)


@dataclass
class DeduplicationResult(Generic[Record]):
    """
    Deduplication result.

    Attributes
    ----------
        selected: List of deduplicated records after removing duplicates.
        filtered: List of DuplicateRecord objects containing details about duplicates of an original record.
        threshold: The similarity threshold used for deduplication.
        columns: Columns used for deduplication.
        deduplicated: Deprecated, use selected instead.
        duplicates: Deprecated, use filtered instead.

    """

    selected: list[Record] = field(default_factory=list)
    filtered: list[DuplicateRecord] = field(default_factory=list)
    threshold: float = field(default=0.9)
    columns: Sequence[str] | None = field(default=None)
    deduplicated: list[Record] = field(default_factory=list)  # Deprecated
    duplicates: list[DuplicateRecord] = field(default_factory=list)  # Deprecated

    def __post_init__(self) -> None:
        """Initialize deprecated fields and warn about deprecation."""
        if self.deduplicated or self.duplicates:
            warnings.warn(
                "'deduplicated' and 'duplicates' fields are deprecated and will be removed in a future release. Use 'selected' and 'filtered' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        if not self.selected and self.deduplicated:
            self.selected = self.deduplicated
        if not self.filtered and self.duplicates:
            self.filtered = self.duplicates
        if not self.deduplicated:
            self.deduplicated = self.selected
        if not self.duplicates:
            self.duplicates = self.filtered

    @property
    def duplicate_ratio(self) -> float:
        """Return the percentage of records dropped."""
        if denom := len(self.selected) + len(self.filtered):
            return 1.0 - len(self.selected) / denom
        return 0.0

    @property
    def exact_duplicate_ratio(self) -> float:
        """Return the percentage of records dropped due to an exact match."""
        if denom := len(self.selected) + len(self.filtered):
            return len([dup for dup in self.filtered if dup.exact]) / denom
        return 0.0

    def get_least_similar_from_duplicates(self, n: int = 1) -> list[tuple[Record, Record, float]]:
        """
        Return the N least similar duplicate pairs.

        :param n: The number of least similar pairs to return.
        :return: A list of tuples consisting of (original_record, duplicate_record, score).
        """
        all_pairs = [(dup.record, d, score) for dup in self.filtered for d, score in dup.duplicates]
        sorted_pairs = sorted(all_pairs, key=lambda x: x[2])  # Sort by score
        return sorted_pairs[:n]

    def rethreshold(self, threshold: float) -> None:
        """Rethreshold the duplicates."""
        if self.threshold > threshold:
            raise ValueError("Threshold is smaller than the given value.")
        for dup in self.filtered:
            dup._rethreshold(threshold)
            if not dup.duplicates:
                self.filtered.remove(dup)
                self.selected.append(dup.record)
        self.threshold = threshold

    @property
    def selected_with_duplicates(self) -> list[SelectedWithDuplicates[Record]]:
        """
        For every kept record, return the duplicates that were removed along with their similarity scores.

        :return: A list of tuples where each tuple contains a kept record
                and a list of its duplicates with their similarity scores.
        """

        def _to_hashable(record: Record) -> frozendict[str, str] | str:
            """Convert a record to a hashable representation."""
            if isinstance(record, dict) and self.columns is not None:
                # Convert dict to frozendict for immutability and hashability
                return to_frozendict(record, set(self.columns))
            return str(record)

        # Build a mapping from original-record  to  [(duplicate, score), …]
        buckets: defaultdict[Hashable, DuplicateList] = defaultdict(list)
        for duplicate_record in self.filtered:
            for original_record, score in duplicate_record.duplicates:
                buckets[_to_hashable(original_record)].append((duplicate_record.record, float(score)))

        result: list[SelectedWithDuplicates[Record]] = []
        for selected in self.selected:
            # Get the list of duplicates for the selected record
            raw_list = buckets.get(_to_hashable(selected), [])
            # Ensure we don't have duplicates in the list
            # Use full-record canonical JSON for dicts so that unhashable values are handled correctly
            deduped = {
                (
                    json.dumps(rec, sort_keys=True, separators=(",", ":"), ensure_ascii=False)
                    if isinstance(rec, dict)
                    else rec
                ): (rec, score)
                for rec, score in raw_list
            }
            result.append(SelectedWithDuplicates(record=selected, duplicates=list(deduped.values())))

        return result


@dataclass
class FilterResult(Generic[Record]):
    """
    Result of filtering operations.

    Attributes
    ----------
        selected: List of records that passed the filter criteria.
        filtered: List of records that were filtered out.
        scores_selected: List of scores for the selected records.
        scores_filtered: List of scores for the filtered records.

    """

    selected: list[Record]
    filtered: list[Record]
    scores_selected: list[float] = field(default_factory=list)
    scores_filtered: list[float] = field(default_factory=list)

    @property
    def filter_ratio(self) -> float:
        """Return the percentage of records filtered out."""
        if denom := len(self.selected) + len(self.filtered):
            return len(self.filtered) / denom
        return 0.0

    @property
    def selected_ratio(self) -> float:
        """Return the percentage of records selected."""
        return 1 - self.filter_ratio



from typing import Any, Protocol, Sequence, Union

import numpy as np
from frozendict import frozendict


class Encoder(Protocol):
    """An encoder protocol for SemHash."""

    def encode(
        self,
        sentences: Union[list[str], str, Sequence[str]],
        **kwargs: Any,
    ) -> np.ndarray:
        """
        Encode a list of sentences into embeddings.

        :param sentences: A list of sentences to encode.
        :param **kwargs: Additional keyword arguments.
        :return: The embeddings of the sentences.
        """
        ...  # pragma: no cover


def to_frozendict(record: dict[str, str], columns: set[str]) -> frozendict[str, str]:
    """Convert a record to a frozendict."""
    return frozendict({k: record.get(k, "") for k in columns})


def compute_candidate_limit(
    total: int,
    selection_size: int,
    fraction: float = 0.1,
    min_candidates: int = 100,
    max_candidates: int = 1000,
) -> int:
    """
    Compute the 'auto' candidate limit based on the total number of records.

    :param total: Total number of records.
    :param selection_size: Number of representatives to select.
    :param fraction: Fraction of total records to consider as candidates.
    :param min_candidates: Minimum number of candidates.
    :param max_candidates: Maximum number of candidates.
    :return: Computed candidate limit.
    """
    # 1) fraction of total
    limit = int(total * fraction)
    # 2) ensure enough to pick selection_size
    limit = max(limit, selection_size)
    # 3) enforce lower bound
    limit = max(limit, min_candidates)
    # 4) enforce upper bound (and never exceed the dataset)
    limit = min(limit, max_candidates, total)
    return limit



from typing import Sequence

from semhash.datamodels import DeduplicationResult, DuplicateRecord


def dict_to_string(record: dict[str, str], columns: Sequence[str]) -> str:
    r"""
    Turn a record into a single string.

    Uses self.columns to determine the order of the text segments.
    Each text is cleaned by replacing '\t' with ' '. The texts are then joined by '\t'.

    :param record: A record to unpack.
    :param columns: Columns to unpack.
    :return: A single string representation of the record.
    """
    return "\t".join(record.get(c, "").replace("\t", " ") for c in columns)


def map_deduplication_result_to_strings(result: DeduplicationResult, columns: Sequence[str]) -> DeduplicationResult:
    """Convert the record and duplicates in each DuplicateRecord back to strings if self.was_string is True."""
    deduplicated_str = [dict_to_string(r, columns) for r in result.selected]
    mapped = []
    for dup_rec in result.duplicates:
        record_as_str = dict_to_string(dup_rec.record, columns)
        duplicates_as_str = [(dict_to_string(r, columns), score) for r, score in dup_rec.duplicates]
        mapped.append(
            DuplicateRecord(
                record=record_as_str,
                duplicates=duplicates_as_str,
                exact=dup_rec.exact,
            )
        )
    return DeduplicationResult(selected=deduplicated_str, filtered=mapped, threshold=result.threshold)


def add_scores_to_records(records: list[dict[str, str]]) -> list[tuple[dict[str, str], float]]:
    """Add scores to records and return a DeduplicationResult."""
    return [(record, 1.0) for record in records]



from __future__ import annotations

from collections import defaultdict
from math import ceil
from typing import Any, Generic, Literal, Sequence

import numpy as np
from frozendict import frozendict
from model2vec import StaticModel
from vicinity import Backend

from semhash.datamodels import DeduplicationResult, DuplicateRecord, FilterResult, Record
from semhash.index import Index
from semhash.records import add_scores_to_records, map_deduplication_result_to_strings
from semhash.utils import Encoder, compute_candidate_limit, to_frozendict


class SemHash(Generic[Record]):
    def __init__(self, index: Index, model: Encoder, columns: Sequence[str], was_string: bool) -> None:
        """
        Initialize SemHash.

        :param index: An index.
        :param model: A model to use for featurization.
        :param columns: Columns of the records.
        :param was_string: Whether the records were strings. Used for mapping back to strings.
        """
        self.index = index
        self.model = model
        self.columns = columns
        self._was_string = was_string
        self._ranking_cache: FilterResult | None = None

    @staticmethod
    def _featurize(
        records: Sequence[dict[str, str]],
        columns: Sequence[str],
        model: Encoder,
    ) -> np.ndarray:
        """
        Featurize a list of records using the model.

        :param records: A list of records.
        :param columns: Columns to featurize.
        :param model: An Encoder model.
        :return: The embeddings of the records.
        """
        # Extract the embeddings for each column across all records
        embeddings_per_col = []
        for col in columns:
            col_texts = [r[col] for r in records]
            col_emb = model.encode(col_texts)
            embeddings_per_col.append(np.asarray(col_emb))

        return np.concatenate(embeddings_per_col, axis=1)

    @classmethod
    def _remove_exact_duplicates(
        cls,
        records: Sequence[dict[str, str]],
        columns: Sequence[str],
        reference_records: list[list[dict[str, str]]] | None = None,
    ) -> tuple[list[dict[str, str]], list[tuple[dict[str, str], list[dict[str, str]]]]]:
        """
        Remove exact duplicates based on the unpacked string representation of each record.

        If reference_records is None, the function will only check for duplicates within the records list.

        :param records: A list of records to check for exact duplicates.
        :param columns: Columns to unpack.
        :param reference_records: A list of records to compare against. These are already unpacked
        :return: A list of deduplicated records and a list of duplicates.
        """
        deduplicated = []
        duplicates = []

        column_set = set(columns)
        # Build a seen set from reference_records if provided
        seen: defaultdict[frozendict[str, str], list[dict[str, str]]] = defaultdict(list)
        if reference_records is not None:
            for record_set in reference_records:
                key = to_frozendict(record_set[0], column_set)
                seen[key] = list(record_set)
        in_one_set = reference_records is None

        for record in records:
            frozen_record = frozendict({k: v for k, v in record.items() if k in column_set})
            if duplicated_records := seen.get(frozen_record):
                duplicates.append((record, duplicated_records))
            else:
                deduplicated.append(record)
                # Only add current documents to seen if no reference set is used
                if in_one_set:
                    seen[frozen_record].append(record)

        return deduplicated, duplicates

    @classmethod
    def from_records(
        cls,
        records: Sequence[Record],
        columns: Sequence[str] | None = None,
        use_ann: bool = True,
        model: Encoder | None = None,
        ann_backend: Backend | str = Backend.USEARCH,
        **kwargs: Any,
    ) -> SemHash:
        """
        Initialize a SemHash instance from records.

        This removes exact duplicates, featurizes the records, and fits a vicinity index.

        :param records: A list of records (strings or dictionaries).
        :param columns: Columns to featurize if records are dictionaries.
        :param use_ann: Whether to use approximate nearest neighbors (True) or basic search (False). Default is True.
        :param model: (Optional) An Encoder model. If None, the default model is used (minishlab/potion-base-8M).
        :param ann_backend: (Optional) The ANN backend to use if use_ann is True. Defaults to Backend.USEARCH.
        :param **kwargs: Any additional keyword arguments to pass to the Vicinity index.
        :return: A SemHash instance with a fitted vicinity index.
        :raises ValueError: If columns are not provided for dictionary records.
        """
        if columns is None and isinstance(records[0], dict):
            raise ValueError("Columns must be specified when passing dictionaries.")

        if isinstance(records[0], str):
            # If records are strings, convert to dictionaries with a single column
            columns = ["text"]
            dict_records: list[dict[str, str]] = [{"text": record} for record in records]
            was_string = True
        else:
            dict_records = list(records)
            was_string = False

        # If no model is provided, load the default model
        if model is None:
            model = StaticModel.from_pretrained("minishlab/potion-base-8M")

        # Remove exact duplicates
        deduplicated_records, duplicates = cls._remove_exact_duplicates(dict_records, columns)

        col_set = set(columns)
        duplicate_map = defaultdict(list)
        for x, _ in duplicates:
            frozen_record = to_frozendict(x, col_set)
            duplicate_map[frozen_record].append(x)

        items: list[list[dict[str, str]]] = []
        for record in deduplicated_records:
            i = [record]
            frozen_record = to_frozendict(record, set(columns))
            i.extend(duplicate_map[frozen_record])
            items.append(i)

        # Create embeddings and unpack records
        embeddings = cls._featurize(deduplicated_records, columns, model)

        # Build the Vicinity index
        backend = ann_backend if use_ann else Backend.BASIC
        index = Index.from_vectors_and_items(
            vectors=embeddings,
            items=items,
            backend_type=backend,
            **kwargs,
        )

        return cls(index=index, columns=columns, model=model, was_string=was_string)

    def deduplicate(
        self,
        records: Sequence[Record],
        threshold: float = 0.9,
    ) -> DeduplicationResult:
        """
        Perform deduplication against the fitted index.

        This method assumes you have already fit on a reference dataset (e.g., a train set) with from_records.
        It will remove any items from 'records' that are similar above a certain threshold
        to any item in the fitted dataset.

        :param records: A new set of records (e.g., test set) to deduplicate against the fitted dataset.
        :param threshold: Similarity threshold for deduplication.
        :return: A deduplicated list of records.
        """
        dict_records = self._validate_if_strings(records)

        # Remove exact duplicates before embedding
        dict_records, exact_duplicates = self._remove_exact_duplicates(
            records=dict_records, columns=self.columns, reference_records=self.index.items
        )
        duplicate_records = []
        for record, duplicates in exact_duplicates:
            duplicated_with_score = add_scores_to_records(duplicates)
            duplicate_record = DuplicateRecord(record=record, duplicates=duplicated_with_score, exact=True)
            duplicate_records.append(duplicate_record)

        # If no records are left after removing exact duplicates, return early
        if not dict_records:
            return DeduplicationResult(
                deduplicated=[], duplicates=duplicate_records, threshold=threshold, columns=self.columns
            )

        # Compute embeddings for the new records
        embeddings = self._featurize(records=dict_records, columns=self.columns, model=self.model)
        # Query the fitted index
        results = self.index.query_threshold(embeddings, threshold=threshold)

        deduplicated_records = []
        for record, similar_items in zip(dict_records, results):
            if not similar_items:
                # No duplicates found, keep this record
                deduplicated_records.append(record)
            else:
                duplicate_records.append(
                    DuplicateRecord(
                        record=record,
                        duplicates=[(item, score) for item, score in similar_items],
                        exact=False,
                    )
                )

        result = DeduplicationResult(
            deduplicated=deduplicated_records, duplicates=duplicate_records, threshold=threshold, columns=self.columns
        )

        if self._was_string:
            # Convert records back to strings if the records were originally strings
            return map_deduplication_result_to_strings(result, columns=self.columns)

        return result

    def self_deduplicate(
        self,
        threshold: float = 0.9,
    ) -> DeduplicationResult:
        """
        Deduplicate within the same dataset. This can be used to remove duplicates from a single dataset.

        :param threshold: Similarity threshold for deduplication.
        :return: A deduplicated list of records.
        """
        # Query the fitted index
        results = self.index.query_threshold(self.index.vectors, threshold=threshold)
        column_set = set(self.columns)

        duplicate_records = []

        deduplicated_records = []
        seen_items: set[frozendict[str, str]] = set()
        for item, similar_items in zip(self.index.items, results):
            # Items is a list of items which are exact duplicates of each other
            # So if the an item has more than one record, it is an exact duplicate
            # Crucially, we should count each instance separately.
            record, *duplicates = item
            # We need to compare all duplicates to all _items_.
            # The first item in a list of duplicate is not duplicated, because otherwise
            # we would remove the whole cluster. But it is a duplicate for the other items.

            # Iterate from index 1.
            for index, curr_record in enumerate(duplicates, 1):
                # The use of indexing is intentional here, we want to check if the object is the same
                # not if they have the same values. If we did != or is we would probably ignore lots
                # of items.
                items_to_keep = item[:index] + item[index + 1 :]
                items_with_score = add_scores_to_records(items_to_keep)
                duplicate_records.append(DuplicateRecord(record=curr_record, duplicates=items_with_score, exact=True))

            # If we don't see any similar_items, we know the record is not a duplicate.
            # in rare cases, the item itself might not be a duplicate of itself.
            if not similar_items:
                deduplicated_records.append(record)
                continue
            items, _ = zip(*similar_items)
            frozen_items = [to_frozendict(item, column_set) for item in items]
            # similar_items includes 'record' itself
            # If we've seen any of these items before, this is a duplicate cluster.
            if any(item in seen_items for item in frozen_items):
                duplicate_records.append(
                    DuplicateRecord(
                        record=record,
                        duplicates=[(item, score) for item, score in similar_items if item != record],
                        exact=False,
                    )
                )
                continue
            # This is the first time we see this cluster of similar items
            deduplicated_records.append(record)
            # Mark all items in this cluster as seen
            seen_items.update(frozen_items)

        result = DeduplicationResult(
            deduplicated=deduplicated_records, duplicates=duplicate_records, threshold=threshold, columns=self.columns
        )

        if self._was_string:
            # Convert records back to strings if the records were originally strings
            return map_deduplication_result_to_strings(result, columns=self.columns)

        return result

    def _validate_if_strings(self, records: Sequence[dict[str, str] | str]) -> Sequence[dict[str, str]]:
        """
        Validate if the records are strings.

        If the records are strings, they are converted to dictionaries with a single column.

        :param records: The records to validate.
        :return: The records as a list of dictionaries.
        :raises ValueError: If the records are strings but were not originally strings.
        :raises ValueError: If the records are not all strings or dictionaries.
        """
        if isinstance(records[0], str):
            if not self._was_string:
                raise ValueError("Records were not originally strings, but you passed strings.")
            dict_records = [{"text": record} for record in records if isinstance(record, str)]
        else:
            dict_records = [record for record in records if isinstance(record, dict)]
        if len(dict_records) != len(records):
            raise ValueError("Records must be either strings or dictionaries.")
        return dict_records

    def find_representative(
        self,
        records: Sequence[Record],
        selection_size: int = 10,
        candidate_limit: int | Literal["auto"] = "auto",
        lambda_param: float = 0.5,
    ) -> FilterResult:
        """
        Find representative samples from a given set of records against the fitted index.

        First, the records are ranked using average similarity.
        Then, the top candidates are re-ranked using Maximal Marginal Relevance (MMR)
        to select a diverse set of representatives.

        :param records: The records to rank and select representatives from.
        :param selection_size: Number of representatives to select.
        :param candidate_limit: Number of top candidates to consider for MMR reranking.
            Defaults to "auto", which calculates the limit based on the total number of records.
        :param lambda_param: Trade-off parameter between relevance (1.0) and diversity (0.0). Must be between 0 and 1.
        :return: A FilterResult with the diversified candidates.
        """
        ranking = self._rank_by_average_similarity(records)
        if candidate_limit == "auto":
            candidate_limit = compute_candidate_limit(total=len(ranking.selected), selection_size=selection_size)
        return self._mmr(ranking, candidate_limit, selection_size, lambda_param)

    def self_find_representative(
        self,
        selection_size: int = 10,
        candidate_limit: int | Literal["auto"] = "auto",
        lambda_param: float = 0.5,
    ) -> FilterResult:
        """
        Find representative samples from the fitted dataset.

        First, the rank the records are ranked using average similarity.
        Then, the top candidates are re-ranked using Maximal Marginal Relevance (MMR)
        to select a diverse set of representatives.

        :param selection_size: Number of representatives to select.
        :param candidate_limit: Number of top candidates to consider for MMR reranking.
            Defaults to "auto", which calculates the limit based on the total number of records.
        :param lambda_param: Trade-off parameter between relevance (1.0) and diversity (0.0). Must be between 0 and 1.
        :return: A FilterResult with the diversified representatives.
        """
        ranking = self._self_rank_by_average_similarity()
        if candidate_limit == "auto":
            candidate_limit = compute_candidate_limit(total=len(ranking.selected), selection_size=selection_size)
        return self._mmr(ranking, candidate_limit, selection_size, lambda_param)

    def filter_outliers(
        self,
        records: Sequence[Record],
        outlier_percentage: float = 0.1,
    ) -> FilterResult:
        """
        Filter outliers in a given set of records against the fitted dataset.

        This method ranks the records by their average similarity and filters the bottom
        outlier_percentage of records as outliers.

        :param records: A sequence of records to find outliers in.
        :param outlier_percentage: The percentage (between 0 and 1) of records to consider outliers.
        :return: A FilterResult where 'selected' contains the inliers and 'filtered' contains the outliers.
        :raises ValueError: If outlier_percentage is not between 0 and 1.
        """
        if outlier_percentage < 0.0 or outlier_percentage > 1.0:
            raise ValueError("outlier_percentage must be between 0 and 1")
        ranking = self._rank_by_average_similarity(records)
        outlier_count = ceil(len(ranking.selected) * outlier_percentage)
        if outlier_count == 0:
            # If the outlier count is 0, return an empty selection
            return FilterResult(
                selected=[], filtered=ranking.selected, scores_selected=[], scores_filtered=ranking.scores_selected
            )

        outlier_records = ranking.selected[-outlier_count:]
        outlier_scores = ranking.scores_selected[-outlier_count:]
        inlier_records = ranking.selected[:-outlier_count]
        inlier_scores = ranking.scores_selected[:-outlier_count]

        return FilterResult(
            selected=inlier_records,
            filtered=outlier_records,
            scores_selected=inlier_scores,
            scores_filtered=outlier_scores,
        )

    def self_filter_outliers(
        self,
        outlier_percentage: float = 0.1,
    ) -> FilterResult:
        """
        Filter outliers in the fitted dataset.

        The method ranks the records stored in the index and filters the bottom outlier_percentage
        of records as outliers.

        :param outlier_percentage: The percentage (between 0 and 1) of records to consider as outliers.
        :return: A FilterResult where 'selected' contains the inliers and 'filtered' contains the outliers.
        :raises ValueError: If outlier_percentage is not between 0 and 1.
        """
        if outlier_percentage < 0.0 or outlier_percentage > 1.0:
            raise ValueError("outlier_percentage must be between 0 and 1")
        ranking = self._self_rank_by_average_similarity()
        outlier_count = ceil(len(ranking.selected) * outlier_percentage)
        if outlier_count == 0:
            # If the outlier count is 0, return an empty selection
            return FilterResult(
                selected=[], filtered=ranking.selected, scores_selected=[], scores_filtered=ranking.scores_selected
            )

        outlier_records = ranking.selected[-outlier_count:]
        outlier_scores = ranking.scores_selected[-outlier_count:]
        inlier_records = ranking.selected[:-outlier_count]
        inlier_scores = ranking.scores_selected[:-outlier_count]

        return FilterResult(
            selected=inlier_records,
            filtered=outlier_records,
            scores_selected=inlier_scores,
            scores_filtered=outlier_scores,
        )

    def _rank_by_average_similarity(
        self,
        records: Sequence[Record],
    ) -> FilterResult:
        """
        Rank a given set of records based on the average cosine similarity of the neighbors in the fitted index.

        :param records: A sequence of records.
        :return: A FilterResult containing the ranking (records sorted and their average similarity scores).
        """
        dict_records = self._validate_if_strings(records)
        embeddings = self._featurize(records=dict_records, columns=self.columns, model=self.model)
        results = self.index.query_top_k(embeddings, k=100, vectors_are_in_index=False)

        # Compute the average similarity for each record.
        sorted_scores = sorted(
            ((record, np.mean(sims)) for record, (_, sims) in zip(dict_records, results)),
            key=lambda x: x[1],
            reverse=True,
        )
        selected, scores_selected = zip(*sorted_scores)

        return FilterResult(
            selected=list(selected),
            filtered=[],
            scores_selected=list(scores_selected),
            scores_filtered=[],
        )

    def _self_rank_by_average_similarity(
        self,
    ) -> FilterResult:
        """
        Rank the records stored in the fitted index based on the average cosine similarity of the neighbors.

        :return: A FilterResult containing the ranking.
        """
        if self._ranking_cache is not None:
            return self._ranking_cache

        dict_records = [record[0] for record in self.index.items]
        results = self.index.query_top_k(self.index.vectors, k=100, vectors_are_in_index=True)

        # Compute the average similarity for each record.
        sorted_scores = sorted(
            ((record, np.mean(sims)) for record, (_, sims) in zip(dict_records, results)),
            key=lambda x: x[1],
            reverse=True,
        )
        selected, scores_selected = zip(*sorted_scores)

        ranking = FilterResult(
            selected=list(selected),
            filtered=[],
            scores_selected=list(scores_selected),
            scores_filtered=[],
        )
        self._ranking_cache = ranking
        return ranking

    def _mmr(
        self,
        ranked_results: FilterResult,
        candidate_limit: int,
        selection_size: int,
        lambda_param: float,
    ) -> FilterResult:
        """
        Perform Maximal Marginal Relevance (MMR) re-ranking on the top candidates from a FilterResult.

        This function first slices the ranking, then computes embeddings for the candidates,
        normalizes them, and finally performs MMR re-ranking to obtain a diverse subset of representatives.

        :param ranked_results: A FilterResult containing the ranking of records.
        :param candidate_limit: Number of top candidates to consider from the ranking.
        :param selection_size: Number of final candidates to select using MMR.
        :param lambda_param: Weight balancing relevance and diversity (between 0 and 1).
        :return: A FilterResult containing the selected (diversified) representatives along with
                their MMR scores in `scores_selected`. The remaining candidates are placed in filtered.
        :raises ValueError: If lambda_param is not between 0 and 1.
        """
        if not (0.0 <= lambda_param <= 1.0):
            raise ValueError("lambda_param must be between 0 and 1")

        # Slice the top candidates from the ranking.
        candidate_records = ranked_results.selected[:candidate_limit]
        candidate_relevance = ranked_results.scores_selected[:candidate_limit]

        # Compute embeddings for candidate records.
        embeddings = self._featurize(records=candidate_records, columns=self.columns, model=self.model)

        # Normalize embeddings for cosine similarity.
        normalized_embeddings = embeddings / (np.linalg.norm(embeddings, axis=1, keepdims=True) + 1e-16)

        # Package candidates as tuples: (record, baseline_relevance, normalized_embedding)
        candidates = [
            (record, relevance, normalized_embedding)
            for record, relevance, normalized_embedding in zip(
                candidate_records, candidate_relevance, normalized_embeddings
            )
        ]

        # If no candidates, return empty result.
        if not candidates:
            return FilterResult(selected=[], filtered=[], scores_selected=[], scores_filtered=[])

        # Initialize selected set with the most relevant candidate.
        first_record, first_relevance, first_embedding = candidates[0]
        selected_records = [first_record]
        selected_scores = [first_relevance]
        selected_embeddings = [first_embedding]
        remaining_candidates = candidates[1:]

        # Iteratively select candidates using the MMR criterion.
        while remaining_candidates and len(selected_records) < selection_size:
            # Build arrays for the remaining candidates.
            embeddings_remaining = np.vstack([emb for (_, _, emb) in remaining_candidates])
            relevances_remaining = np.array([rel for (_, rel, _) in remaining_candidates])

            # Build array of embeddings for the already selected set.
            embeddings_selected = np.vstack(selected_embeddings)

            # Compute cosine similarities between selected and remaining candidates.
            similarity_matrix = embeddings_remaining.dot(embeddings_selected.T)
            max_similarity = similarity_matrix.max(axis=1)

            # Compute MMR scores for all remaining candidates.
            mmr_scores = lambda_param * relevances_remaining - (1.0 - lambda_param) * max_similarity

            # Choose the candidate with the highest MMR score.
            best_index = int(np.argmax(mmr_scores))
            record, _, embedding = remaining_candidates.pop(best_index)

            selected_records.append(record)
            selected_scores.append(mmr_scores[best_index])
            selected_embeddings.append(embedding)

        # Whatever is left in remaining_candidates is filtered out.
        filtered_records = [rec for (rec, _, _) in remaining_candidates]
        filtered_scores = [rel for (_, rel, _) in remaining_candidates]

        return FilterResult(
            selected=selected_records,
            filtered=filtered_records,
            scores_selected=selected_scores,
            scores_filtered=filtered_scores,
        )



from dataclasses import dataclass


@dataclass
class DatasetRecord:
    """Dataset record."""

    name: str
    text_name: str | None = None
    label_name: str | None = None
    sub_directory: str = ""
    columns: list[str] | None = None
    split_one: str = "train"
    split_two: str = "test"


DATASET_DICT: dict[str, DatasetRecord] = {
    "bbc": DatasetRecord(name="SetFit/bbc-news", text_name="text", label_name="label_text"),
    "senteval_cr": DatasetRecord(name="SetFit/SentEval-CR", text_name="text", label_name="label_text"),
    "tweet_sentiment_extraction": DatasetRecord(
        name="SetFit/tweet_sentiment_extraction", text_name="text", label_name="label_text"
    ),
    "emotion": DatasetRecord(name="SetFit/emotion", text_name="text", label_name="label_text"),
    "amazon_counterfactual": DatasetRecord(
        name="SetFit/amazon_counterfactual_en", text_name="text", label_name="label_text"
    ),
    "ag_news": DatasetRecord(name="SetFit/ag_news", text_name="text", label_name="label_text"),
    "enron_spam": DatasetRecord(name="SetFit/enron_spam", text_name="text", label_name="label_text"),
    "subj": DatasetRecord(name="SetFit/subj", text_name="text", label_name="label_text"),
    "sst5": DatasetRecord(name="SetFit/sst5", text_name="text", label_name="label_text"),
    "20_newgroups": DatasetRecord(name="SetFit/20_newsgroups", text_name="text", label_name="label_text"),
    "hatespeech_offensive": DatasetRecord(name="SetFit/hate_speech_offensive", text_name="text", label_name="label"),
    "ade": DatasetRecord(name="SetFit/ade_corpus_v2_classification", text_name="text", label_name="label"),
    "imdb": DatasetRecord(name="SetFit/imdb", text_name="text", label_name="label"),
    "massive_scenario": DatasetRecord(
        name="SetFit/amazon_massive_scenario_en-US", text_name="text", label_name="label"
    ),
    "student": DatasetRecord(name="SetFit/student-question-categories", text_name="text", label_name="label"),
    "squad_v2": DatasetRecord(name="squad_v2", columns=["question", "context"], split_two="validation"),
    "wikitext": DatasetRecord(
        name="Salesforce/wikitext", text_name="text", label_name="text", sub_directory="wikitext-103-raw-v1"
    ),
}



from __future__ import annotations

from typing import Any

import numpy as np
from vicinity import Backend
from vicinity.backends import AbstractBackend, get_backend_class
from vicinity.datatypes import SingleQueryResult

DocScore = tuple[dict[str, str], float]
DocScores = list[DocScore]
DictItem = list[dict[str, str]]


class Index:
    def __init__(self, vectors: np.ndarray, items: list[DictItem], backend: AbstractBackend) -> None:
        """
        An index that maps vectors to items.

        This index has an efficient backend for querying, but also explicitly stores the vectors in memory.

        :param vectors: The vectors of the items.
        :param items: The items in the index. This is a list of lists. Each sublist contains one or more dictionaries
            that represent records. These records are exact duplicates of each other.
        :param backend: The backend to use for querying.
        """
        self.items = items
        self.backend = backend
        self.vectors = vectors

    @classmethod
    def from_vectors_and_items(
        cls, vectors: np.ndarray, items: list[DictItem], backend_type: Backend | str, **kwargs: Any
    ) -> Index:
        """
        Load the index from vectors and items.

        :param vectors: The vectors of the items.
        :param items: The items in the index.
        :param backend_type: The type of backend to use.
        :param **kwargs: Additional arguments to pass to the backend.
        :return: The index.
        """
        backend_class = get_backend_class(backend_type)
        arguments = backend_class.argument_class(**kwargs)
        backend = backend_class.from_vectors(vectors, **arguments.dict())

        return cls(vectors, items, backend)

    def query_threshold(self, vectors: np.ndarray, threshold: float) -> list[DocScores]:
        """
        Query the index with a threshold.

        :param vectors: The vectors to query.
        :param threshold: The similarity threshold.
        :return: The query results.
        """
        out: list[DocScores] = []
        for result in self.backend.threshold(vectors, threshold=1 - threshold, max_k=100):
            intermediate = []
            for index, distance in zip(*result):
                # Every item in the index contains one or more records.
                # These are all exact duplicates, so they get the same score.
                for record in self.items[index]:
                    # The score is the cosine similarity.
                    # The backend returns distances, so we need to convert.
                    intermediate.append((record, 1 - distance))
            out.append(intermediate)

        return out

    def query_top_k(self, vectors: np.ndarray, k: int, vectors_are_in_index: bool) -> list[SingleQueryResult]:
        """
        Query the index with a top-k.

        :param vectors: The vectors to query.
        :param k: Maximum number of top-k records to keep.
        :param vectors_are_in_index: Whether the vectors are in the index. If this is set to True, we retrieve k + 1
            records, and do not consider the first one, as it is the query vector itself.
        :return: The query results. Each result is a tuple where the first element is the list of neighbor records,
                 and the second element is a NumPy array of cosine similarity scores.
        """
        results = []
        offset = int(vectors_are_in_index)
        for x, y in self.backend.query(vectors=vectors, k=k + offset):
            # Convert returned distances to cosine similarities.
            similarities = 1 - y[offset:]
            results.append((x[offset:], similarities))
        return results
